#  text comparator
=begin
    Copyright (C) 2008  Tom Gregorovi, _tom_ at centrum.cz

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
=end

require 'cgi'
require 'iconv'
require 'open-uri'
require 'strscan'
require 'htmlobjects'

def UnicodetoUTF8(code)
  if code < 0x000080 then
    return code.chr
  else
    if code < 0x0007FF then
      return (((code & 0x7C0) >> 6) | 0xC0).chr + ((code & 0x3F) | 0x80).chr
    else
      if code < 0x010000 then
        return (((code & 0xF000) >> 12) | 0xE0).chr + (((code & 0xFC0) >> 6) | 0x3F).chr + ((code & 0x3F) | 0x80).chr
      else
        return (((code & 0x1C0000) >> 18) | 0xF0).chr + (((code & 0x3F000) >> 12) | 0x3F).chr + (((code & 0xFC0) >> 6) | 0x3F).chr + ((code & 0x3F) | 0x80).chr
      end
    end
  end
end

class HTMLPage
  def initialize
    @html = ''
    @words = []
  end
  def html
    @html
  end
  def words
    @words
  end
  def downloadtime
    @downloadtime
  end
  def parsetime
    @parsetime
  end
  # regular expression matching html tags and comments
  def tagregex 
    /<((!--.*?--[^>]*?)|(([^>]*?(("[">]*?")|('[^']*?'))*?)*?))>/m
  end
  # parse html page from url into array of words
  def parse(url)
    charset = 'ISO-8859-1'
    
    @downloadtime = Time.now
    
    open(url) {|f|
      @html = f.read
      if f.charset != nil then
        charset = f.charset.upcase
      end 
    }
    
    @parsetime = Time.now
    @downloadtime = @parsetime - @downloadtime
      
    # get html source charset
    case @html[0, 2] # check for UTF-16 Byte Order Mask
    when "\xFE\xFF"
      charset = 'UTF-16BE'
    when "\xFF\xFE"
      charset = 'UTF-16LE'  
    end
    
    if @html[1] == "\x00" then
      charset = 'UTF-16BE'
    end
    if @html[0] == "\x00" then
      charset = 'UTF-16LE'
    end
    
    if charset[0, 6] == 'UTF-16' then
      @html = Iconv.iconv('UTF-8', charset, @html).to_s
      charset = 'UTF-8'
    end
    
    scanner = StringScanner.new(@html)
    scannerupcase = StringScanner.new(@html.upcase)
    textscanner = StringScanner.new('')
    
    # parse header
    while not scannerupcase.eos?
      scannerupcase.skip(/[^<]+/m)
      
      tag = scannerupcase.scan(tagregex)
      if tag == nil then
        if scanner.skip(/</) != nil then 
          next
        else
          break
        end
      end
      
      tagpart = tag[1, 4]
      if tagpart == 'META' then
        textscanner.string = tag
        ch = textscanner.skip_until(/charset/i)
        if ch != nil then
          r = textscanner.skip(/\s*=\s*/i)
          if r != nil then
            c = textscanner.scan(/[^"']+/).delete("\"'")
            if c != nil then
              charset = c
            end  
          end  
        end
      end
      if tagpart == 'BODY' then break end
    end
    
    if charset != 'UTF-8' then
      @html = Iconv.iconv('UTF-8', charset, @html).to_s
      scanner = StringScanner.new(@html)
      scannerupcase = StringScanner.new(@html.upcase)
    end
    
    # parse body to text
    skip = 0
    @newword = true
    
    def addtext(text)
      if text != nil then
        if text.length > 0 then
          if text[0].chr =~ /(\s)/ then
            @newword = true
          else
            if @newword or @words.length == 0 then
              @words.push(text)
              @newword = false
            else
              @words[@words.length - 1] = @words.at(@words.length - 1) + text
            end
          end
        end
      end
    end
    
    while not scannerupcase.eos?
      scanner.pos = scannerupcase.pos
      t = scanner.scan(/[^<]+/)
      if t != nil then
        if skip == 0 then
          # parse HTML entities
          textscanner.string = t
          
          while not textscanner.eos?
            s = textscanner.skip(/\s+/)
            if s != nil then
              @newword = true
            end
            
            s = textscanner.scan(/([^&\s])+/)
            addtext(s)
            
            e = textscanner.skip(/&/)
            if e != nil then
              e = textscanner.scan(/[#\w]+?;/)
              if e != nil then
                if e[0].chr == '#' then # entity is unicode char code
                  v = ('0' + e[1, e.length - 2]).to_i
                  u = UnicodetoUTF8(v)
                else # named entity
                  u = HTMLEntities[e[0, e.length - 1]]
                  if u == nil then u = '&' + e end
                end
                addtext(u)
              else
                addtext('&')
              end
            end
          end
        end
      end
      
      scannerupcase.pos = scanner.pos
      tag = scannerupcase.scan(tagregex)
      if tag == nil then
        if scannerupcase.skip(/</) != nil then
          if skip == 0 then
            addtext('<')
          end
          next
        else
          break
        end
      else
        textscanner.string = tag
        tagname = textscanner.scan(/<([\/])?([^\/\s>])+/)
        if tagname != nil then
          close = (tagname[1].chr == '/')
          if not close then
            tagname = tagname[1, tagname.length - 1]
          else
            tagname = tagname[2, tagname.length - 2]
          end
          
          # check for skip
          if HTMLSkipTags.key?(tagname) then
            if not close then
              skip = skip + 1
            else
              if skip > 0 then skip = skip - 1 end
            end
          else
            # block tags split words
            if HTMLBlockTags.key?(tagname) and skip == 0 then @newword = true end
          end
        end
      end
    end
    
    @parsetime = Time.now - @parsetime
  end
  
  # convert html source to simple html code preserving block tags (like H1, PRE, DIV) while stripping style
  # words selected in mask will be surrounded by selection tag
  def getsimplehtml(selectionmask = [], selectiontag = 'i', source = @html)
    result = ''
    skip = 0
    @selectionmask = selectionmask
    @selectiontag = selectiontag
    
    def addtext(text, code)
      if text != nil then
        if text.length > 0 then
          if text[0].chr =~ /(\s)/ then
            @newword = true
            
            return code
          else
            if @newword then
              @wordindex = @wordindex + 1
              @newword = false
            end
            
            if @selectionmask[@wordindex] > 0 then
              return '<' + @selectiontag + '>' + code + '</' + @selectiontag + '>'
            else
              return code
            end
          end
        end
      end
      
      return ''
    end
    
    scanner = StringScanner.new(source)
    scannerupcase = StringScanner.new(source.upcase)
    textscanner = StringScanner.new('')
    @newword = true
    @wordindex = -1
    
    while not scannerupcase.eos?
      scanner.pos = scannerupcase.pos
      t = scanner.scan(/[^<]+/)
      if t != nil then # use text as is
        if skip == 0 then 
          textscanner.string = t
          
          while not textscanner.eos?
            s = textscanner.scan(/\s+/)
            if s != nil then
              @newword = true
              result = result + s
            end
            
            s = textscanner.scan(/([^&\s])+/)
            result = result + addtext(s, s)
            
            e = textscanner.skip(/&/)
            if e != nil then
              e = textscanner.scan(/[#\w]+?;/)
              if e != nil then
                if e[0].chr == '#' then # entity is unicode char code
                  v = ('0' + e[1, e.length - 2]).to_i
                  u = UnicodetoUTF8(v)
                else # named entity
                  u = HTMLEntities[e[0, e.length - 1]]
                  if u == nil then u = '&' + e end
                end
                result = result + addtext(u, '&' + e)
              else
                result = result + addtext('&', '&')
              end
            end
          end
        end
      end
      
      scannerupcase.pos = scanner.pos
      tag = scannerupcase.scan(tagregex)
      if tag == nil then
        if scannerupcase.skip(/</) != nil then
          if skip == 0 then result = result + addtext('<', '<') end
          next
        else
          break
        end
      else
        textscanner.string = tag
        tagname = textscanner.scan(/<([\/])?([^\/\s>])+/)
        if tagname != nil then
          close = (tagname[1].chr == '/')
          if not close then
            tagname = tagname[1, tagname.length - 1]
          else
            tagname = tagname[2, tagname.length - 2]
          end
          
          # check for skip
          if HTMLSkipTags.key?(tagname) then
            if not close then
              skip = skip + 1
            else
              if skip > 0 then skip = skip - 1 end
            end
          else
            # check block tags
            if HTMLBlockTags.key?(tagname) and skip == 0 then
              @newword = true
              
              if not close then
                result = result + '<' + tagname + '>'
              else
                result = result + '</' + tagname + '>'
              end
            end
          end
        end
      end
    end
    
    return result
  end
end