# plagiarism search application class
=begin
    Copyright (C) 2008  Tomáš Gregorovič, _tom_ at centrum.cz

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
=end

require 'textparser'
require 'htmltotext'
require 'ysearch'
require 'thread'

class PlagiarismItem
  def initialize(url, htmlurl, title, summary, mimetype)
    @url = url
    @htmlurl = htmlurl
    @title = title
    @summary = summary
    @mimetype = mimetype
    @error = ''
    @checked = false
  end
  def checked?
    @checked
  end
  def download
    @page = HTMLPage.new
    @page.parse(@htmlurl)
  end
  def check(pattern)
    @document = ParsedDocument.new(pattern, @page.words)
    @checked = true
  end
  def url
    @url
  end
  def htmlurl
    @htmlurl
  end
  def title
    @title
  end
  def mimetype
    @mimetype
  end
  def page
    @page
  end
  def document
    @document
  end
  def summary
    @summary
  end
  def setsummary(summary)
    if summary.similarityindex > @summary.similarityindex then
      @summary = summary
    end
  end
  def error
    @error
  end
  def seterror(error)
    @error = error
    @checked = true
  end
  def itemtype
    if @error != '' then
      :error
    else
      if (summary.similarwordscount >= 3) and (summary.similarityratio > 0.2) then
        if document == nil then 
          :normalresult
        else
          if (document.similarwordscount >= 4) and (document.similarityratio > 0.3) then
            :normal
          else
            :hidden
          end
        end
      else
        :hiddenresult
      end
    end
  end
end

class Plagiarism
  def initialize(pattern)
    @results = {}
    @resultlist = []
    @documentlist = []
    @pattern = pattern
    @searchfrom = 0
    @resultcount = 0
    @nonewresult = false
  end
  def newstep
    @searchtime = 0
    @processtime = 0
    @laststepresults = {:normalresult => 0, :hiddenresult => 0}
    @laststepdocuments = {:normal => 0, :hidden => 0, :error => 0}
  end
  def laststepresults
    @laststepresults
  end
  def laststepdocuments
    @laststepdocuments
  end
  # hash for urls => PlagirismItem
  def results
    @results
  end
  # sorted by summary similarity index
  def resultlist
    @resultlist
  end
  # sorted by document similarity index
  def documentlist
    @documentlist
  end
  def searchfrom
    if @pattern.queries.length > 0 then
      @pattern.queries[0].from
	else
	  0
	end
  end
  def searchcount
    if @pattern.queries.length > 0 then
      @pattern.queries[0].requestedcount
	else
	  0
	end
  end
  def resultcount
    @resultcount
  end
  def pattern
    @pattern
  end
  def searchtime
    @searchtime
  end
  def processtime
    @processtime
  end
  def find(similar = 0, from = 0)
    if @pattern.queries.length == 0 then
	  return
	end
	
    @searchtime = Time.now
    threads = Array.new
    @searchfrom = from
	
	# recount max search query results by last successes
	maxcount = 250
	s = 0
	@pattern.queries.each { |query|
      if not query.exhausted? then
	    s = s + query.foundcount + 1
      end
	}
    
    if s == 0 then
      @searchtime = Time.now - @searchtime
      return
    end
    
    @pattern.queries.each { |query|
	  if not query.exhausted? then
        threads << Thread.new {
	      query.setfrom(query.from + query.requestedcount)
          
          l = query.foundcount + 1
          if s < 1 then s = 1 end
          f = (l.to_f * maxcount.to_f / s.to_f).round
          if f < 1 then f = 1 end
          
          s = s - l
          maxcount = maxcount - f
          
		  query.setrequestedcount([100, f].min) 
          searcher = WebSearch.new('hWID2RfV34GfWHNpU4Qe8LPSVAUFAod.I3bjhvQkw671jMrYrKvpKazqSI7B2Q',
            query.words.join(' '), query.type, query.requestedcount, query.from, nil, 1)
		   
          searcher.set_similar_ok(similar)
       
          resultset = searcher.parse_results()
          Thread.current[:result] = resultset
          query.setreturnedcount(resultset.length)
          Thread.current[:query] = query
        }
      else
        query.setrequestedcount(0)
      end
    }
    
    threads.each { |thread|
      if thread.join(10) != nil and thread[:result] != nil then
        # get result title + summary similarity with query 
        query = thread[:query]
        
		c = 0
        thread[:result].each { |r|
          summary = ParsedSummary.new(query, r['Summary'])
          
          item = @results[r['Url']]
          if item == nil then
            if r['MimeType'] == 'text/html' then
              item = PlagiarismItem.new(r['Url'], r['Url'], r['Title'], summary, r['MimeType'])
              @results[r['Url']] = item
              @laststepresults[item.itemtype] = @laststepresults[item.itemtype] + 1
            end
          else
            if not item.checked? then
              item.setsummary(summary)
            end
          end
		  if item != nil and item.itemtype != :hiddenresult then
		    c = c + 1
		  end
        }
		
		query.setfoundcount(c)
      end   
    }
    
    # sort not checked items by summary similarity index
    @resultlist = @results.values.sort { |a, b|
      if a.checked? then
        if b.checked? then
          -(a.summary.similarityindex <=> b.summary.similarityindex)
        else
          -1
        end
      else
        if b.checked? then
          if a.checked? then
            -(a.summary.similarityindex <=> b.summary.similarityindex)
          else
            1
          end
        else
          -(a.summary.similarityindex <=> b.summary.similarityindex)
        end
      end
    }    
    
    @searchtime = Time.now - @searchtime
  end
  def checksimilarity(count)
    @processtime = Time.now
    threads = Array.new
    
    checklist = []
    
    @resultlist.each_index { |i| item = @resultlist[i]
      if not item.checked? and item.itemtype != :hiddenresult then
        checklist.push(item)
        @resultcount = @resultcount + 1
        if @resultcount >= count then break end
      end
    }
    
    checklist.each { |item|
      threads << Thread.new {
        begin
          Thread.current[:url] = item.url
          item.download
        rescue
          item.seterror('Chyba: ' + $!.to_s + ' ' + $!.backtrace.join(' '))
        end
      } 
    }

	list = []
	t = Time.now
	while Time.now - t < 10 
      threads.each { |thread|
        if thread.join(0.1) != nil then
          list.push(@results[thread[:url]])
		  threads.delete(thread)
        end
      }
	end
	
	threads.each { |thread|
        if thread.join(4) == nil then
          @results[thread[:url]].seterror('Chyba: Vypršel čas pro zpracování dotazu.')
        end
        list.push(@results[thread[:url]])
      }
	
	list.each { |item|
	  item.check(@pattern)
      @laststepdocuments[item.itemtype] = @laststepdocuments[item.itemtype] + 1
	}
    
    @documentlist.concat(checklist)
    
    @documentlist.sort! { |a, b|
      if a.error != '' then
        1
      else
        if b.error != '' then
          -1
        else
          -(a.document.similarityindex <=> b.document.similarityindex )
        end
      end
    }
    
    @processtime = Time.now - @processtime
  end
end