/*
    Copyright (C) 2008  Tom Gregorovi, _tom_ at centrum.cz

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Include the Ruby headers and goodies
#include "ruby.h"

// Defining a space for information and references about the module to be stored internally
VALUE TextSimilarity = Qnil;

// Prototype for the initialization method - Ruby calls this, not you
void Init_textsimilarity();

VALUE method_getwordsimilarity(VALUE self, VALUE allwordmatches);

// The initialization method for this module
void Init_textsimilarity() {
	TextSimilarity = rb_define_module("TextSimilarity");
	rb_define_method(TextSimilarity, "getwordsimilarity", method_getwordsimilarity, 1);
}

/* faster version - use this if you don't need to treat negative offset */
static inline VALUE
rb_ary_elt(ary, offset)
    VALUE ary;
    long offset;
{
    if (RARRAY(ary)->len == 0) return Qnil;
    if (offset < 0 || RARRAY(ary)->len <= offset) {
	return Qnil;
    }
    return RARRAY(ary)->ptr[offset];
}

// counts similarity index of each word
VALUE method_getwordsimilarity(VALUE self, VALUE allwordmatches) {
    
	long len = RARRAY(allwordmatches)->len;
	long *array = (long *)malloc(len * sizeof(long));
	
	long maxdistance = 5;
	// weights for distance between two same words
	long similaritytable[] = {28, 24, 21, 17, 14, 12, 10, 8, 3, 1};
	long similaritytablelen = 10;
	long minwordsimilarity = 40;
	
    long i;
	for (i = 0; i < len; i++) {
	  array[i] = 0;
	  
	  VALUE posi = rb_ary_elt(allwordmatches, i);
	  if (posi == Qnil) {continue;}
	 
	  long maxwordsimilarityindex = 0;
	  
      long i1;
	  for (i1 = 0; i1 < RARRAY(posi)->len; i1++) {        
        long wordsimilarityindex = 0;
        
        long j;
        for (j = i - maxdistance; j <= i + maxdistance; j++) {
	      long dij = i - j;
	      if (dij == 0) {continue;}
          
          VALUE posj = rb_ary_elt(allwordmatches, j);
	      if (posj == Qnil) {continue;}
        
          long minindex = similaritytablelen;
          
		  long j1;
		  for (j1 = 0; j1 < RARRAY(posj)->len; j1++) {
		    long posi1 = RARRAY(posi)->ptr[i1];
	        long posj1 = RARRAY(posj)->ptr[j1];
			long dpos = posi1 - posj1;
			if (dpos == 0) {continue;}
			
			long s = abs(dij) + abs(dpos) - 2;
            if (dij * dpos < 0) {
              s = s + 1;
            }
            if (s < minindex) {
              minindex = s;
            }
		  }
          
          if (minindex < similaritytablelen) {
            wordsimilarityindex = wordsimilarityindex + similaritytable[minindex];
          }
		}
        if (wordsimilarityindex > maxwordsimilarityindex) {
          maxwordsimilarityindex = wordsimilarityindex;
        }
      }
	  
	  if (maxwordsimilarityindex > minwordsimilarity) {
        if (maxwordsimilarityindex > 100) {
          array[i] = 100;
        } else {
          array[i] = maxwordsimilarityindex;
        }
	  }
	}
	
	
	// convert c array to ruby array
	VALUE result = rb_ary_new2(len);
	for (i = 0; i < len; i++) {
	    rb_ary_store(result, i, rb_float_new(array[i] / 100.0));
	}
	
	return result;
}
